@@ -1,74 +1,59 @@
 {-# LANGUAGE CPP #-}
+{-# LANGUAGE OverloadedStrings #-}
 {-# LANGUAGE ScopedTypeVariables #-}
+{-# LANGUAGE TupleSections #-}
+{-# OPTIONS_GHC -fno-warn-deprecations #-}
 
 module Network.Wai.Handler.Warp.Run where
 
 import Control.Concurrent (threadDelay, forkIOWithUnmask)
-import Control.Exception
+import qualified Control.Concurrent as Conc (yield)
+import Control.Exception as E
 import Control.Monad (forever, when, unless, void)
-import Control.Monad.IO.Class (MonadIO, liftIO)
-import Control.Monad.Trans.Class (lift)
 import Data.ByteString (ByteString)
 import qualified Data.ByteString as S
-import Data.Conduit
-import Data.Conduit.Internal (ResumableSource (..))
-import qualified Data.Conduit.List as CL
-import Data.Conduit.Network (bindPort)
+import Data.Streaming.Network (bindPortTCP)
 import Network (sClose, Socket)
-import Network.Sendfile
-import Network.Socket (accept, SockAddr)
+import Network.Socket (accept, withSocketsDo, SockAddr)
 import qualified Network.Socket.ByteString as Sock
 import Network.Wai
+import Network.Wai.Internal (ResponseReceived (ResponseReceived))
+import qualified Network.Wai.Handler.Warp.Date as D
+import qualified Network.Wai.Handler.Warp.FdCache as F
+import Network.Wai.Handler.Warp.Buffer
+import Network.Wai.Handler.Warp.Header
+import Network.Wai.Handler.Warp.Recv
 import Network.Wai.Handler.Warp.Request
 import Network.Wai.Handler.Warp.Response
+import Network.Wai.Handler.Warp.SendFile
 import Network.Wai.Handler.Warp.Settings
 import qualified Network.Wai.Handler.Warp.Timeout as T
 import Network.Wai.Handler.Warp.Types
-import Prelude hiding (catch)
-
--- Sock.recv first tries to call recvfrom() optimistically.
--- If EAGAIN returns, it polls incoming data with epoll/kqueue.
--- This code first polls incoming data with epoll/kqueue.
+import Data.IORef (IORef, newIORef, readIORef, writeIORef)
 
 #if WINDOWS
-import qualified Control.Concurrent.MVar as MV
-import Network.Socket (withSocketsDo)
-import Control.Concurrent (forkIO)
+import Network.Wai.Handler.Warp.Windows
 #else
 import System.Posix.IO (FdOption(CloseOnExec), setFdOption)
 import Network.Socket (fdSocket)
 #endif
 
-#if SENDFILEFD
-import Control.Applicative
-import qualified Network.Wai.Handler.Warp.FdCache as F
-#endif
-
--- FIXME come up with good values here
-bytesPerRead :: Int
-bytesPerRead = 4096
-
--- | Default action value for 'Connection'
-socketConnection :: Socket -> Connection
-socketConnection s = Connection
-    { connSendMany = Sock.sendMany s
-    , connSendAll = Sock.sendAll s
-    , connSendFile = sendFile s
-    , connClose = sClose s
-    , connRecv = Sock.recv s bytesPerRead
-    }
-
-sendFile :: Socket -> FilePath -> Integer -> Integer -> IO () -> [ByteString] -> Cleaner -> IO ()
-#if SENDFILEFD
-sendFile s path off len act hdr cleaner = case fdCacher cleaner of
-    Nothing  -> sendfileWithHeader s path (PartOfFile off len) act hdr
-    Just fdc -> do
-        (fd, fresher) <- F.getFd fdc path
-        sendfileFdWithHeader s fd (PartOfFile off len) (act>>fresher) hdr
-#else
-sendFile s path off len act hdr _ =
-    sendfileWithHeader s path (PartOfFile off len) act hdr
-#endif
+-- | Default action value for 'Connection'.
+socketConnection :: Socket -> IO Connection
+socketConnection s = do
+    readBuf <- allocateBuffer bufferSize
+    writeBuf <- allocateBuffer bufferSize
+    return Connection {
+        connSendMany = Sock.sendMany s
+      , connSendAll = Sock.sendAll s
+      , connSendFile = defaultSendFile s
+      , connClose = sClose s >> freeBuffer readBuf >> freeBuffer writeBuf
+      , connRecv = receive s readBuf bufferSize
+      , connReadBuffer = readBuf
+      , connWriteBuffer = writeBuf
+      , connBufferSize = bufferSize
+      , connSendFileOverride = Override s
+      }
 
 #if __GLASGOW_HASKELL__ < 702
 allowInterrupt :: IO ()
@@ -80,28 +65,15 @@ allowInterrupt = unblock $ return ()
 run :: Port -> Application -> IO ()
 run p = runSettings defaultSettings { settingsPort = p }
 
--- | Run a Warp server with the given settings.
+-- | Run an 'Application' with the given 'Settings'.
 runSettings :: Settings -> Application -> IO ()
-#if WINDOWS
-runSettings set app = withSocketsDo $ do
-    var <- MV.newMVar Nothing
-    let clean = MV.modifyMVar_ var $ \s -> maybe (return ()) sClose s >> return Nothing
-    void . forkIO $ bracket
-        (bindPort (settingsPort set) (settingsHost set))
-        (const clean)
-        (\s -> do
-            MV.modifyMVar_ var (\_ -> return $ Just s)
-            runSettingsSocket set s app)
-    forever (threadDelay maxBound) `finally` clean
-#else
-runSettings set app =
+runSettings set app = withSocketsDo $
     bracket
-        (bindPort (settingsPort set) (settingsHost set))
+        (bindPortTCP (settingsPort set) (settingsHost set))
         sClose
         (\socket -> do
             setSocketCloseOnExec socket
             runSettingsSocket set socket app)
-#endif
 
 -- | Same as 'runSettings', but uses a user-supplied socket instead of opening
 -- one. This allows the user to provide, for example, Unix named socket, which
@@ -111,13 +83,24 @@ runSettings set app =
 -- 'serverPort' record.
 runSettingsSocket :: Settings -> Socket -> Application -> IO ()
 runSettingsSocket set socket app =
-    runSettingsConnection set getter app
+    runSettingsConnection set getConn app
   where
-    getter = do
-        (conn, sa) <- accept socket
-        setSocketCloseOnExec socket
-        return (socketConnection conn, sa)
+    getConn = do
+#if WINDOWS
+        (s, sa) <- windowsThreadBlockHack $ accept socket
+#else
+        (s, sa) <- accept socket
+#endif
+        setSocketCloseOnExec s
+        conn <- socketConnection s
+        return (conn, sa)
 
+-- | Allows you to provide a function which will return a 'Connection'. In
+-- cases where creating the @Connection@ can be expensive, this allows the
+-- expensive computations to be performed in a separate thread instead of the
+-- main server loop.
+--
+-- Since 1.3.5
 runSettingsConnection :: Settings -> IO (Connection, SockAddr) -> Application -> IO ()
 runSettingsConnection set getConn app = runSettingsConnectionMaker set getConnMaker app
   where
@@ -125,20 +108,18 @@ runSettingsConnection set getConn app = runSettingsConnectionMaker set getConnMa
       (conn, sa) <- getConn
       return (return conn, sa)
 
--- | Allows you to provide a function which will return a @Connection@. In
--- cases where creating the @Connection@ can be expensive, this allows the
--- expensive computations to be performed in a separate thread instead of the
--- main server loop.
---
--- Since 1.3.5
 runSettingsConnectionMaker :: Settings -> IO (IO Connection, SockAddr) -> Application -> IO ()
-runSettingsConnectionMaker set getConn app = do
-#if SENDFILEFD
-    let duration = settingsFdCacheDuration set
-    fc <- case duration of
-        0 -> return Nothing
-        _ -> Just <$> F.initialize (duration * 1000000)
-#endif
+runSettingsConnectionMaker x y =
+    runSettingsConnectionMakerSecure x (go y)
+  where
+    go = fmap (\(a, b) -> (fmap (, False) a, b))
+
+-- | Allows you to provide a function which will return a function
+-- which will return 'Connection'.
+--
+-- Since 2.1.4
+runSettingsConnectionMakerSecure :: Settings -> IO (IO (Connection, Bool), SockAddr) -> Application -> IO ()
+runSettingsConnectionMakerSecure set getConnMaker app = do
     settingsBeforeMainLoop set
 
     -- Note that there is a thorough discussion of the exception safety of the
@@ -157,6 +138,8 @@ runSettingsConnectionMaker set getConn app = do
     -- First mask all exceptions in the main loop. This is necessary to ensure
     -- that no async exception is throw between the call to getConnLoop and the
     -- registering of connClose.
+    D.withDateCache $ \dc -> do
+    F.withFdCache (settingsFdCacheDuration set * 1000000) $ \fc -> do
     withTimeoutManager $ \tm -> mask_ . forever $ do
         -- Allow async exceptions before receiving the next connection maker.
         allowInterrupt
@@ -184,39 +167,39 @@ runSettingsConnectionMaker set getConn app = do
             -- We grab the connection before registering timeouts since the
             -- timeouts will be useless during connection creation, due to the
             -- fact that async exceptions are still masked.
-            bracket mkConn connClose $ \conn ->
+            bracket mkConn (connClose . fst) $ \(conn', isSecure') ->
 
             -- We need to register a timeout handler for this thread, and
             -- cancel that handler as soon as we exit.
             bracket (T.registerKillThread tm) T.cancel $ \th ->
-#if SENDFILEFD
-                let cleaner = Cleaner th fc
-#else
-                let cleaner = Cleaner th
-#endif
+                let ii = InternalInfo th fc dc
+                    conn = setSendFile conn' fc
                     -- We now have fully registered a connection close handler
                     -- in the case of all exceptions, so it is safe to one
                     -- again allow async exceptions.
                  in unmask .
                     -- Call the user-supplied on exception code if any
                     -- exceptions are thrown.
-                    handle onE .
+                    handle (onE Nothing) .
 
                     -- Call the user-supplied code for connection open and close events
-                    bracket_ onOpen onClose $
+                    bracket (onOpen addr) (const $ onClose addr) $ \goingon ->
 
                     -- Actually serve this connection.
-                    serveConnection th set cleaner port app conn addr
+                    -- onnClose above ensures the termination of the connection.
+                    when goingon $ serveConnection conn ii addr isSecure' set app
   where
     -- FIXME: only IOEception is caught. What about other exceptions?
-    getConnLoop = getConn `catch` \(e :: IOException) -> do
-        onE (toException e)
+    getConnLoop = getConnMaker `E.catch` \(e :: IOException) -> do
+        onE Nothing (toException e)
         -- "resource exhausted (Too many open files)" may happen by accept().
         -- Wait a second hoping that resource will be available.
         threadDelay 1000000
         getConnLoop
-    onE = settingsOnException set
-    port = settingsPort set
+    onE mreq e =
+        case fromException e of
+            Just (NotEnoughLines []) -> return ()
+            _ -> settingsOnException set mreq e
     onOpen = settingsOnOpen set
     onClose = settingsOnClose set
 
@@ -228,52 +211,84 @@ runSettingsConnectionMaker set getConn app = do
                 f
             Just tm -> f tm
 
-serveConnection :: T.Handle
+serveConnection :: Connection
+                -> InternalInfo
+                -> SockAddr
+                -> Bool -- ^ is secure?
                 -> Settings
-                -> Cleaner
-                -> Port -> Application -> Connection -> SockAddr-> IO ()
-serveConnection timeoutHandle settings cleaner port app conn remoteHost' =
-    runResourceT serveConnection'
+                -> Application
+                -> IO ()
+serveConnection conn ii addr isSecure' settings app = do
+    istatus <- newIORef False
+    src <- mkSource (connSource conn th istatus)
+    recvSendLoop istatus src `E.catch` \e -> do
+        sendErrorResponse istatus e
+        throwIO (e :: SomeException)
+
   where
-    innerRunResourceT
-        | settingsResourceTPerRequest settings = lift . runResourceT
-        | otherwise = id
-    th = threadHandle cleaner
+    th = threadHandle ii
 
-    serveConnection' :: ResourceT IO ()
-    serveConnection' = serveConnection'' $ connSource conn th
+    sendErrorResponse istatus e = do
+        status <- readIORef istatus
+        when status $ void $
+            sendResponse conn ii dummyreq defaultIndexRequestHeader (return S.empty) (errorResponse e)
 
-    serveConnection'' fromClient = do
-        (env, getSource) <- parseRequestInternal conn timeoutHandle port remoteHost' fromClient
-        case settingsIntercept settings env of
-            Nothing -> do
-                -- Let the application run for as long as it wants
-                liftIO $ T.pause th
-                keepAlive <- innerRunResourceT $ do
-                    res <- app env
+    dummyreq = defaultRequest { remoteHost = addr }
 
-                    liftIO $ T.resume th
-                    sendResponse settings cleaner env conn res
+    errorResponse e = settingsOnExceptionResponse settings e
 
-                -- flush the rest of the request body
-                requestBody env $$ CL.sinkNull
-                ResumableSource fromClient' _ <- liftIO getSource
+    recvSendLoop istatus fromClient = do
+        (req', idxhdr) <- recvRequest settings conn ii addr fromClient
+        let req = req' { isSecure = isSecure' }
+        -- Let the application run for as long as it wants
+        T.pause th
 
-                when keepAlive $ serveConnection'' fromClient'
-            Just intercept -> do
-                liftIO $ T.pause th
-                ResumableSource fromClient' _ <- liftIO getSource
-                intercept fromClient' conn
+        -- In the event that some scarce resource was acquired during
+        -- creating the request, we need to make sure that we don't get
+        -- an async exception before calling the ResponseSource.
+        keepAliveRef <- newIORef $ error "keepAliveRef not filled"
+        _ <- app req $ \res -> do
+            T.resume th
+            -- FIXME consider forcing evaluation of the res here to
+            -- send more meaningful error messages to the user.
+            -- However, it may affect performance.
+            writeIORef istatus False
+            keepAlive <- sendResponse conn ii req idxhdr (readSource fromClient) res
+            writeIORef keepAliveRef keepAlive
+            return ResponseReceived
+        keepAlive <- readIORef keepAliveRef
 
-connSource :: Connection -> T.Handle -> Source (ResourceT IO) ByteString
-connSource Connection { connRecv = recv } th = src
+        -- We just send a Response and it takes a time to
+        -- receive a Request again. If we immediately call recv,
+        -- it is likely to fail and the IO manager works.
+        -- It is very costy. So, we yield to another Haskell
+        -- thread hoping that the next Request will arraive
+        -- when this Haskell thread will be re-scheduled.
+        -- This improves performance at least when
+        -- the number of cores is small.
+        Conc.yield
+
+        when keepAlive $ do
+            -- flush the rest of the request body
+            flushBody $ requestBody req
+            T.resume th
+            recvSendLoop istatus fromClient
+
+flushBody :: IO ByteString -> IO ()
+flushBody src =
+    loop
   where
-    src = do
-        bs <- liftIO recv
-        unless (S.null bs) $ do
-            when (S.length bs >= 2048) $ liftIO $ T.tickle th
-            yield bs
-            src
+    loop = do
+        bs <- src
+        unless (S.null bs) loop
+
+connSource :: Connection -> T.Handle -> IORef Bool -> IO ByteString
+connSource Connection { connRecv = recv } th istatus = do
+    bs <- recv
+    unless (S.null bs) $ do
+        writeIORef istatus True
+        when (S.length bs >= 2048) $ T.tickle th
+    return bs
 
 -- Copied from: https://github.com/mzero/plush/blob/master/src/Plush/Server/Warp.hs
 setSocketCloseOnExec :: Socket -> IO ()
